# You probably don't want to read this code, it's a mess.

import random
import math
import pyglet
import const
import util


class GameWindow(pyglet.window.Window):

    def __init__(self):
        super(GameWindow, self).__init__(const.WIDTH, const.HEIGHT,
                                         const.TITLE)
        self.current_window = Intro(self)

    def start_game(self):
        self.current_window = Game(self)


class Intro:

    def __init__(self, window):
        self.window = window
        self.batch = pyglet.graphics.Batch()
        self.intro = pyglet.sprite.Sprite(pyglet.resource.image('intro.png'),
                                          0, 0, batch=self.batch)
        self.window.push_handlers(self.on_draw, self.on_mouse_press)

    def on_draw(self):
        self.window.clear()
        self.batch.draw()

    def on_mouse_press(self, x, y, button, modifiers):
        self.window.pop_handlers()
        self.window.start_game()


class Game:

    def __init__(self, window):
        self.window = window

        self.paused = False
        self.paused_label = None

        self.batch = pyglet.graphics.Batch()
        self.batch2 = pyglet.graphics.Batch()
        
        self.bg = pyglet.sprite.Sprite(pyglet.resource.image('bg.png'), 0, 0,
                                       batch=self.batch, group=const.GROUP_BG)

        self.board = Board(None) # drawing manually, see below
        self.board.add_piece('pawn', [5, 4])
        self.board.add_piece('king', [4, 6])
        self.board.add_piece('pawn', [3, 4])
        self.board.add_piece('knight', [6, 5])
        self.board.add_piece('knight', [2, 5])        

        self.make_buy_images()

        self.mouse_x = 0
        self.mouse_y = 0

        self.ghost_piece = None

        self.square = None

        self.action = const.NOTHING

        self.square_moves = None
        self.last_piece = None
        self.holding_piece = None

        self.guts = 0
        self.kills = 0

        self.guts_label = pyglet.text.Label(str(self.guts), batch=self.batch2,
                                            x=670, y=122, font_size=16)
        self.kills_label = pyglet.text.Label(str(self.kills),
                                             batch=self.batch2,
                                             x=670, y=73, font_size=16)

        self.zombie_manager = ZombieManager(self.batch, self.batch2,
                                            self.board)

        self.window.push_handlers(self.on_draw, self.on_key_press,
                                  self.on_key_release, self.on_mouse_press,
                                  self.on_mouse_release, self.on_mouse_motion,
                                  self.on_mouse_drag)
        pyglet.clock.schedule_interval(self.update, 1.0/const.FPS)

    def make_buy_images(self):
        self.buy_image = {}
        for piece in const.PIECE_NAMES:
            self.buy_image[piece] = pyglet.resource.image(piece + '.png')
        
    def finish(self):
        self.window.pop_handlers()
        pyglet.clock.unschedule(self.update)

    def on_draw(self):
        self.window.clear()
        self.batch.draw()
        # Sigh. Apparently changing groups repeatedly is a bit bugged.
        # Drawing pieces manually, because they have to be in order.
        self.board.draw()
        self.batch2.draw()

    def on_key_press(self, symbol, modifiers):
        if symbol == pyglet.window.key.P:
            if not self.paused: self.pause()
            else: self.unpause()
        elif symbol == pyglet.window.key.R:
            self.finish()
            self.window.start_game()

    def pause(self):
        self.paused = True
        self.paused_label = pyglet.text.Label('Paused', batch=self.batch2,
            x=const.WIDTH / 2, y=const.HEIGHT / 2, font_size=36,
            anchor_x='center', anchor_y='center', color=(0,0,0,255))

    def unpause(self):
        self.paused = False
        self.paused_label.delete()
        self.paused_label = None

    def on_key_release(self, symbol, modifiers):
        pass

    def on_mouse_press(self, x, y, button, modifiers):
        if self.paused: return
        if button == pyglet.window.mouse.LEFT:
            if x < const.BOARD_END_X:
                for piece in self.board.pieces:
                    if (not piece.moving() and
                        util.inside_sprite_box(x, y, piece)):
                        self.ghost_piece = GhostPiece(self.batch2,
                            piece.image, piece.name, x, y, x - piece.x,
                            y - piece.y)
                        self.square = Square(self.batch, self.board, x, y)
                        self.action = const.MOVING
                        self.holding_piece = piece
                        return
            else:
                self.check_buy(x, y)

    def check_buy(self, x, y):
        for piece in const.PIECE_NAMES:
            if (util.inside(x, y, *const.BUY_POS[piece]) and
                self.guts >= const.COST[piece]):
                self.ghost_piece = GhostPiece(self.batch2,
                    self.buy_image[piece], piece, x, y,
                    x - const.BUY_POS[piece][0]-self.buy_image[piece].width/2,
                    y - const.BUY_POS[piece][1])
                self.square = Square(self.batch, self.board, x, y)
                self.action = const.BUYING
                return

    def check_moves(self, x, y):
        for piece in self.board.pieces:
            if not piece.moving() and util.inside_sprite_box(x, y, piece):
                if self.last_piece != piece:
                    self.last_piece = piece
                    return piece.moves()
                return [None]
        return None

    def on_mouse_release(self, x, y, button, modifiers):
        if self.action == const.BUYING:
            pos = util.board_tile(*self.ghost_piece.position())
            if (util.in_board(pos) and
                self.guts >= const.COST[self.ghost_piece.name]):
                self.board.add_piece(self.ghost_piece.name, pos)
                self.guts -= const.COST[self.ghost_piece.name]
            self.action = const.NOTHING
        elif self.action == const.MOVING:
            pos = util.board_tile(*self.ghost_piece.position())
            if util.in_board(pos) and self.in_square_moves(pos):
                self.board.move(self.holding_piece.pos, pos)
            self.holding_piece = None
            self.last_piece = None
            self.action = const.NOTHING
        if self.ghost_piece:
            self.ghost_piece.delete()
            self.ghost_piece = None
        if self.square:
            self.square.delete()
            self.square = None

    def on_mouse_motion(self, x, y, dx, dy):
        self.mouse_x = x
        self.mouse_y = y

    def hover_moves(self):
        if self.holding_piece: return
        x = self.mouse_x
        y = self.mouse_y
        moves = self.check_moves(x, y)
        if moves != None:
            if moves != [None]:
                self.square_moves = []
                for pos in moves:
                    self.square_moves.append(MovesSquare(self.batch,
                                                         self.board, pos))
        else:
            self.last_piece = None
            if self.square_moves != None:
                for move in self.square_moves:
                    move.delete()
                self.square_moves = None

    def in_square_moves(self, pos):
        if self.square_moves == None: return False
        for move in self.square_moves:
            if pos == move.pos: return True
        return False

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        self.mouse_x = x
        self.mouse_y = y        

    def update(self, dt):
        #print self.mouse_x, self.mouse_y
        if self.ghost_piece:
            self.ghost_piece.update(self.mouse_x, self.mouse_y)
            if self.square:
                pos = self.ghost_piece.position()
                self.square.update(pos[0], pos[1], self.square_moves)
        if self.paused: return

        self.board.update(dt)
        if self.board.game_over:
            self.game_over()
        self.zombie_manager.update(dt)
        self.check_collision()
        
        self.hover_moves()
        self.guts_label.text = str(self.guts)
        self.kills_label.text = str(self.kills)

    def game_over(self):
        self.gameover_label = pyglet.text.Label('Game over',
            batch=self.batch2, x=const.WIDTH / 2, y=const.HEIGHT / 2,
            font_size=36, anchor_x='center', anchor_y='center',
            color=(0,0,0,255))
        self.paused = True

    def check_collision(self):
        for piece in self.board.pieces:
            if piece.moving():
                for zombie in self.zombie_manager.zombies:
                    if util.hit_sprite_h2(piece, zombie):
                        zombie.died()
                        self.kills += 1
                        self.guts += 1
        for zombie in self.zombie_manager.zombies:
            if not zombie.attacking:
                for piece in self.board.pieces:
                    if util.hit(piece, zombie.x, zombie.y + zombie.range):
                        zombie.attack(piece)


class ChessPiece(pyglet.sprite.Sprite):

    def __init__(self, batch, board, pos, name, health=1):
        super(ChessPiece, self).__init__(pyglet.resource.image(name + '.png'),
                                         batch=batch, group=const.GROUP_PIECE)

        self.board = board

        self.name = name
        self.health_max = health
        self.health = self.health_max
        self.images = [self.image]
        for i in range(health):
            self.images.append(
                pyglet.resource.image(name + '-hurt' + str(i+1) + '.png'))
        self.images.append(pyglet.resource.image(name + '-dead.png'))
        for image in self.images: image.anchor_x = image.width / 2

        self.move_to(pos)

        self.target = None
        self.direction = None

        self.speed = 300

    def calculate_position(self):
        self.x = self.pos[0] * const.TILE_SIZE + const.OFFSET_X
        self.y = self.pos[1] * const.TILE_SIZE + const.OFFSET_Y

    def set_position(self, pos):
        self.board.move(self.pos, pos)
        self.pos = pos

    def hurt(self):
        if self.health >= 0:
            self.health -= 1
            self.image = self.images[self.health_max - self.health]
            if self.health < 0:
                self.died()

    def died(self):
        self.opacity = 128
        self.board.remove(self.pos)
        self.board.dead_pieces.append(self)

    def move_to(self, pos):
        self.pos = pos
        self.calculate_position()

    def run_to(self, pos):
        dx = pos[0] - self.pos[0]
        dy = pos[1] - self.pos[1]
        d = math.hypot(dx, dy)
        if d == 0: return
        dx /= d
        dy /= d
        self.direction = [dx, dy]
        self.target = [pos[0] * const.TILE_SIZE + const.OFFSET_X,
                       pos[1] * const.TILE_SIZE + const.OFFSET_Y]
        self.pos = pos # think about this

    def moving(self):
        return (self.target != None)

    def moves(self):
        pass

    def update(self, dt):
        if self.target and self.direction:
            ok = 0
            if abs(self.x - self.target[0]) <= dt * self.speed + 1:
                self.x = self.target[0]
                ok += 1
            if abs(self.y - self.target[1]) <= dt * self.speed + 1:
                self.y = self.target[1]
                ok += 1
            if ok == 2:
                self.target = None
                self.direction = None
                self.calculate_position()
            else:
                self.x += self.direction[0] * dt * self.speed
                self.y += self.direction[1] * dt * self.speed
                


class GhostPiece(pyglet.sprite.Sprite):

    def __init__(self, batch, image, name, x, y, anchor_x, anchor_y):
        super(GhostPiece, self).__init__(image, batch=batch,
                                         group=const.GROUP_GHOST)
        # there should be a better way to copy an image, no?
        self.image = self.image.get_texture().get_region(0, 0,
                         self.image.width, self.image.height)
        self.image.anchor_x = self.image.width / 2 + anchor_x
        self.image.anchor_y = anchor_y
        self.anc_x = anchor_x
        self.anc_y = anchor_y
        self.x = x
        self.y = y
        self.opacity = 128
        self.name = name

    def update(self, mouse_x, mouse_y):
        self.x = mouse_x
        self.y = mouse_y

    def position(self):
        return [self.x - self.anc_x, self.y - self.anc_y]


class Pawn(ChessPiece):

    def __init__(self, batch, board, pos):
        super(Pawn, self).__init__(batch, board, pos, 'pawn')

    def moves(self):
        ps = [[self.pos[0], self.pos[1] + 1],
              [self.pos[0], self.pos[1] - 1]]
        moves = [p for p in ps if self.board.free(p)]
        return moves


class Knight(ChessPiece):

    def __init__(self, batch, board, pos):
        super(Knight, self).__init__(batch, board, pos, 'knight')

    def moves(self):
        ps = [[self.pos[0] + 1, self.pos[1] + 2],
              [self.pos[0] + 1, self.pos[1] - 2],
              [self.pos[0] - 1, self.pos[1] + 2],
              [self.pos[0] - 1, self.pos[1] - 2],
              [self.pos[0] + 2, self.pos[1] + 1],
              [self.pos[0] + 2, self.pos[1] - 1],
              [self.pos[0] - 2, self.pos[1] + 1],
              [self.pos[0] - 2, self.pos[1] - 1]]
        moves = [p for p in ps if self.board.free(p)]
        return moves


class Bishop(ChessPiece):

    def __init__(self, batch, board, pos):
        super(Bishop, self).__init__(batch, board, pos, 'bishop')

    def moves(self):
        moves = []
        for i in range(1, 8):
            p = [self.pos[0] + i, self.pos[1] + i]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0] - i, self.pos[1] + i]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0] + i, self.pos[1] - i]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0] - i, self.pos[1] - i]
            if self.board.free(p): moves.append(p)
            else: break
        return moves

class Rook(ChessPiece):

    def __init__(self, batch, board, pos):
        super(Rook, self).__init__(batch, board, pos, 'rook')

    def moves(self):
        moves = []
        for i in range(1, 8):
            p = [self.pos[0] + i, self.pos[1]]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0] - i, self.pos[1]]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0], self.pos[1] + i]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0], self.pos[1] - i]
            if self.board.free(p): moves.append(p)
            else: break
        return moves


class Queen(ChessPiece):

    def __init__(self, batch, board, pos):
        super(Queen, self).__init__(batch, board, pos, 'queen')

    def moves(self):
        moves = []
        for i in range(1, 8):
            p = [self.pos[0] + i, self.pos[1] + i]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0] - i, self.pos[1] + i]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0] + i, self.pos[1] - i]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0] - i, self.pos[1] - i]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0] + i, self.pos[1]]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0] - i, self.pos[1]]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0], self.pos[1] + i]
            if self.board.free(p): moves.append(p)
            else: break
        for i in range(1, 8):
            p = [self.pos[0], self.pos[1] - i]
            if self.board.free(p): moves.append(p)
            else: break
        return moves

class King(ChessPiece):

    def __init__(self, batch, board, pos):
        super(King, self).__init__(batch, board, pos, 'king')

    def moves(self):
        ps = [[self.pos[0] + 1, self.pos[1]],
              [self.pos[0] - 1, self.pos[1]],
              [self.pos[0], self.pos[1] + 1],
              [self.pos[0], self.pos[1] - 1]]
        moves = [p for p in ps if self.board.free(p)]
        return moves


class Board:

    def __init__(self, batch):
        self.board = [[None for j in range(8)] for i in range(8)]
        self.pieces = []
        self.batch = batch
        self.dead_pieces = []
        self.king = None
        self.game_over = False

    def add(self, piece, pos):
        self.board[pos[0]][pos[1]] = piece
        self.pieces.append(piece)

    def add_piece(self, name, pos):
        if pos[0] < 0 or pos[0] >= 8 or pos[1] < 0 or pos[1] >= 8:
            return False
        if name == 'pawn': piece = Pawn(self.batch, self, pos)
        elif name == 'knight': piece = Knight(self.batch, self, pos)
        elif name == 'bishop': piece = Bishop(self.batch, self, pos)
        elif name == 'rook': piece = Rook(self.batch, self, pos)
        elif name == 'queen': piece = Queen(self.batch, self, pos)
        elif name == 'king':
            piece = King(self.batch, self, pos)
            self.king = piece
        else: return False
        self.add(piece, pos)
        return True

    def remove(self, pos):
        if self.board[pos[0]][pos[1]] == self.king:
            self.game_over = True
        self.pieces.remove(self.board[pos[0]][pos[1]])
        self.board[pos[0]][pos[1]] = None

    def get(self, pos):
        return self.board[pos[0]][pos[1]]

    def move(self, fr, to, instant=False):
        if fr != to and self.board[fr[0]][fr[1]] != None:
            if instant:
                self.board[fr[0]][fr[1]].move_to(to)
            else:
                self.board[fr[0]][fr[1]].run_to(to)                
            self.board[to[0]][to[1]] = self.get(fr)
            self.board[fr[0]][fr[1]] = None

    def draw(self):
        for piece in self.dead_pieces:
            piece.draw()
        for i in range(8):
            for j in range(7, -1, -1): # must be drawn in this order
                if self.board[i][j]:
                    self.board[i][j].draw()

    def free(self, pos):
        return util.in_board(pos) and self.board[pos[0]][pos[1]] == None

    def update(self, dt):
        for piece in self.pieces:
            piece.update(dt)
            

class Square(pyglet.sprite.Sprite):

    def __init__(self, batch, board, x, y):
        super(Square, self).__init__(pyglet.resource.image('square.png'),
                                     batch=batch, group=const.GROUP_SQUARE)
        self.op = 128
        self.opacity = self.op
        self.board = board
        self.update(x, y, None)
        self.set_no()

    def set_move(self):
        self.color = (255, 255, 0)

    def set_no(self):
        self.color = (255, 0, 0)

    def set_ok(self):
        self.color = (0, 255, 0)

    def update(self, mouse_x, mouse_y, square_moves):
        self.pos = util.board_tile(mouse_x, mouse_y)
        if (self.pos[0] < 0 or self.pos[0] >= 8 or
            self.pos[1] < 0 or self.pos[1] >= 8):
            if self.opacity != 0: self.opacity = 0
        else:
            if self.opacity != self.op: self.opacity = self.op
            if self.board.get(self.pos) == None:
                if square_moves != None:
                    ok = False
                    for move in square_moves:
                        if self.pos == move.pos:
                            ok = True
                            break
                    if ok: self.set_ok()
                    else: self.set_no()
                else:
                    self.set_ok()
            else:
                self.set_no()
        self.x = self.pos[0] * const.TILE_SIZE + const.BOARD_START_X
        self.y = self.pos[1] * const.TILE_SIZE + const.BOARD_START_Y


class MovesSquare(pyglet.sprite.Sprite):

    def __init__(self, batch, board, pos):
        super(MovesSquare, self).__init__(pyglet.resource.image('square.png'),
                                          batch=batch,
                                          group=const.GROUP_SQUARE)
        self.op = 128
        self.opacity = self.op
        self.board = board
        self.pos = pos
        self.x = self.pos[0] * const.TILE_SIZE + const.BOARD_START_X
        self.y = self.pos[1] * const.TILE_SIZE + const.BOARD_START_Y
        self.color = (0, 0, 255)


class Zombie(pyglet.sprite.Sprite):

    def __init__(self, batch, board, x, y):
        img_full = pyglet.resource.image('zombie1-walk.png')
        img_seq = pyglet.image.ImageGrid(img_full, 1, 4)
        walk_rate = 0.2
        self.walk_anim = pyglet.image.Animation.from_image_sequence(
            img_seq, walk_rate, True)
        for img in img_seq: img.anchor_x = img.width / 2
        self.walk_anim.anchor_x = img_seq[0].anchor_x
        self.walk_anim.anchor_y = img_seq[0].anchor_y

        img_full = pyglet.resource.image('zombie1-attack.png')
        img_seq = pyglet.image.ImageGrid(img_full, 1, 4)
        attack_rate = 0.25
        self.attack_anim = pyglet.image.Animation.from_image_sequence(
            img_seq, attack_rate, True)
        for img in img_seq: img.anchor_x = img.width / 2
        self.attack_anim.anchor_x = img_seq[0].anchor_x
        self.attack_anim.anchor_y = img_seq[0].anchor_y
        
        super(Zombie, self).__init__(self.walk_anim, batch=batch,
                                     group=const.GROUP_ZOMBIE, x=x, y=y)

        self.dead_image = pyglet.resource.image('zombie1-dead.png')

        self.speed = 20
        self.board = board
        self.direction = const.UP

        self.dead = False
        self.timer = 0
        self.remove_me = False

        self.attacking = False
        self.piece_attacking = None
        self.range = 100

    def update(self, dt):
        if not self.attacking:
            if self.direction == const.UP:
                self.y += self.speed * dt
                self.shuffle_x(dt)
                if self.y >= const.BOARD_END_Y:
                    self.board.king.hurt()
                    self.disappeared()
#                    self.direction = const.DOWN
#            if self.direction == const.DOWN:
#                self.y -= self.speed * dt
#                self.shuffle_x(dt)
#                if self.y < const.BOARD_START_Y:
#                    self.direction = const.UP
        else:
            if self.piece_attacking.moving():
                pyglet.clock.unschedule(self.hurt_piece)
                if not self.dead:
                    self.image = self.walk_anim
                else:
                    self.image = self.dead_image
                self.attacking = False
                self.piece_attacking = None

    def attack(self, piece):
        self.image = self.attack_anim
        self.attacking = True
        self.piece_attacking = piece
        pyglet.clock.schedule_once(self.hurt_piece, 1, piece)

    def hurt_piece(self, dt, piece):
        if not self.dead:
            piece.hurt()
            self.image = self.walk_anim
        self.attacking = False
        self.piece_attacking = None

    def update_dead(self, dt):
        self.timer += dt
        if self.timer > 5:
            self.remove_me = True

    def shuffle_x(self, dt):
        if random.uniform(0, 1) < 0.5:
            self.x += random.randint(-20, 20) * dt
            board_x = self.x - const.BOARD_START_X
            target_x = self.board.king.x - const.BOARD_START_X
            if board_x <= target_x:
                if random.uniform(0, 1) > board_x / target_x:
                    self.x += random.randint(0, 50) * dt
            else:
                if (random.uniform(0, 1) < (board_x - target_x) / target_x):
                    self.x -= random.randint(0, 50) * dt

    def died(self):
        self.dead = True
        self.image = self.dead_image
        self.opacity = 192
        pyglet.clock.unschedule(self.hurt_piece)

    def disappeared(self):
        self.remove_me = True


class ZombieManager:

    def __init__(self, batch, batch2, board):
        self.zombies = []
        self.batch = batch
        self.batch2 = batch2
        self.board = board
        self.dead_zombies = []
        self.timer = 0
        self.diff = 0.01
        self.between_time = 3
        self.between_timer = 0

    def add(self, x, y):
        zombie = Zombie(self.batch2, self.board, x, y)
        self.zombies.append(zombie)

    def update(self, dt):
        self.update_zombies(dt)

        self.timer += dt
        if self.timer >= 3:
            self.timer = 0
            self.diff += 0.005
            self.between_time -= 0.1
        self.between_timer -= dt
        if self.between_timer < 0:
            self.between_timer = 0
            
        if self.between_timer == 0 and random.uniform(0, 1) < self.diff:
            self.generate()
            self.between_timer = self.between_time

    def update_zombies(self, dt):
        rm_list = []
        for zombie in self.zombies:
            zombie.update(dt)
            if zombie.dead: rm_list.append(zombie)
        for zombie in rm_list:
            self.zombies.remove(zombie)
            zombie.batch = self.batch
            self.dead_zombies.append(zombie)
        rm_list = []
        for zombie in self.dead_zombies:
            zombie.update_dead(dt)
            if zombie.remove_me: rm_list.append(zombie)
        for zombie in rm_list:
            self.dead_zombies.remove(zombie)
            zombie.delete()

    def generate(self):
        x = random.randint(const.BOARD_START_X, const.BOARD_END_X)
        self.add(x, -20)
            
